/*
    Text maid
    copyright (c) 1998-2008 Kazuki IWAMOTO http://www.maid.org/ iwm@maid.org

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "abort.h"
#include "charset.h"
#include "file.h"
#include "general.h"
#include "prop.h"
#include "sigfile.h"
#include "sigkey.h"
#include "sigmain.h"
#include "misc/charuty.h"
#include "misc/fileio.h"
#include "misc/misc.h"
#include "orz/orzhistory.h"
#include "orz/orzmdi.h"


/******************************************************************************
*                                                                             *
* ja:ファイル入力関数群                                                       *
*                                                                             *
******************************************************************************/
/*  ja:TXTファイルを開く
     file,ファイル名
    tmaid,ウインドウ情報                                                    */
void
file_open_text (const gchar *file,
                TmaidWindow *tmaid)
{
  gchar buf[16384], **charset, *inbuf = NULL, *outbuf = NULL;
  gsize bytes_read, bytes_written;
  gssize leng, length = 0;
  gint i, j;
  gunichar c;
  FileIO *fio;
  LineBuffer *p;
  GError *error = NULL;

  charset = g_strsplit (tmaid->ft.charset, ",", 0);
  g_free (tmaid->ft.charset);
  p = g_malloc0 (sizeof (LineBuffer));
  tmaid->max = 1;
  tmaid->off = 0;
  tmaid->start = p;
  /* ja:中断ボタン */
  abort_box_start ();
  /* ja:読み込み */
  fio = fileio_open (file, FILEIO_ACCESS_READ, FILEIO_SHARE_READ,
                                                    FILEIO_MODE_OPEN_EXISTING);
  if (fio)
    {
      while (abort_break && (leng = fileio_read (fio, buf, 16384)) > 0)
        {
          while (gtk_events_pending ())
            gtk_main_iteration ();
          inbuf = g_realloc (inbuf, length + leng);
          g_memmove (inbuf + length, buf, leng);
          length += leng;
        }
      fileio_close (fio);
    }
  if (!inbuf)
    {
      /* ja:終了処理 */
      /* ja:中断ボタン */
      abort_box_end ();
#ifdef G_OS_WIN32
      tmaid->ft.crlf = 0;
#else /* not G_OS_WIN32 */
      tmaid->ft.crlf = 2;
#endif /* not G_OS_WIN32 */
      tmaid->ft.charset = g_strdup (charset[0]);
      g_strfreev (charset);
      return;
    }
  /* ja:キャラクターセット変換 */
  for (i = 0; abort_break && charset[i]; i++)
    {
      while (gtk_events_pending ())
        gtk_main_iteration ();
      outbuf = g_convert (inbuf, length, "UTF-8", charset[i],
                                                NULL, &bytes_written, &error);
      if (error && g_error_matches (error, G_CONVERT_ERROR,
                                                G_CONVERT_ERROR_NO_CONVERSION))
        {
          /* ja:キャラクターセットが無効 */
          GtkWidget *dialog;

          dialog = gtk_message_dialog_new (GTK_WINDOW (window),
                            GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                            GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
                            _("Conversion from character encoding"
                            " '%s' to 'UTF-8' is not supported"), charset[i]);
          g_signal_connect (G_OBJECT (dialog), "key-press-event",
                                    G_CALLBACK (misc_dialog_key_press), NULL);
          gtk_dialog_run (GTK_DIALOG (dialog));
          gtk_widget_destroy (dialog);
          for (j = 0; j < ftnum; j++)
            {
              gchar *tmp;

              tmp = charuty_remove (ftype[j].charset, charset[i]);
              g_free (ftype[j].charset);
              ftype[j].charset = tmp ? tmp : g_strdup ("UTF-8");
            }
        }
      g_clear_error (&error);
      if (outbuf)
        break;
    }
  /* ja:キャラクターセット変換を強行する */
  if (!outbuf)
    for (i = 0; abort_break && !outbuf && charset[i]; i++)
      {
        gchar *buffer;
        gssize bytes_read_prev = -1;

        buffer = g_memdup (inbuf, length);
        while ((g_clear_error (&error),
                !(outbuf = g_convert (buffer, length, "UTF-8", charset[i],
                                        &bytes_read, &bytes_written, &error)))
                                    && bytes_read_prev != bytes_read && error
                                    && g_error_matches (error, G_CONVERT_ERROR,
                                            G_CONVERT_ERROR_ILLEGAL_SEQUENCE))
          {
            while (gtk_events_pending ())
              gtk_main_iteration ();
            buffer[bytes_read] = '?';
            bytes_read_prev = bytes_read;
          }
        g_free (buffer);
        g_clear_error (&error);
        if (outbuf)
          break;
      }
  /* ja:変換できなければUTF-8とする */
  if (!outbuf)
    {
      gchar *ptr;

      outbuf = inbuf;
      leng = length;
      while (abort_break && leng > 0
                    && !g_utf8_validate (inbuf, leng, (const gchar **)&ptr))
        {
          while (gtk_events_pending ())
            gtk_main_iteration ();
          *ptr = '?';
          leng -= ptr - inbuf;
          inbuf = ptr;
        }
      tmaid->ft.charset = g_strdup ("UTF-8");
    }
  else
    {
      g_free (inbuf);
      length = bytes_written;
      tmaid->ft.charset = g_strdup (charset[i]);
    }
  g_strfreev (charset);
  /* ja:改行コード判定 */
  if (tmaid->ft.crlf >= 3)
    {
      gint crlf[3];

      crlf[0] = crlf[1] = crlf[2] = 0;
      i = 0;
      while (abort_break && i < length)
        if (i <= length - 2 && outbuf[i] == '\r' && outbuf[i + 1] == '\n')
          {
            i += 2;
            crlf[0]++;
          }
        else 
          {
            if (outbuf[i] == '\r')
              crlf[1]++;
            else if (outbuf[i] == '\n')
              crlf[2]++;
            i++;
          }
#ifdef G_OS_WIN32
      if (crlf[0] >= crlf[1] && crlf[0] >= crlf[2])
        tmaid->ft.crlf = 0;
      else if (crlf[2] >= crlf[1])
        tmaid->ft.crlf = 2;
      else
        tmaid->ft.crlf = 1;
#else /* not G_OS_WIN32 */
      if (crlf[2] >= crlf[0] && crlf[2] >= crlf[1])
        tmaid->ft.crlf = 2;
      else if (crlf[0] >= crlf[1])
        tmaid->ft.crlf = 0;
      else
        tmaid->ft.crlf = 1;
#endif /* not G_OS_WIN32 */
    }
  /* ja:ラインバッファに変換する */
  for (i = 0; abort_break && i < length; i = j)
    {
      while (gtk_events_pending ())
        gtk_main_iteration ();
      switch (tmaid->ft.crlf)
        {
          case 0:
            for (j = i; j < length; j++)
              if (j <= length - 2
                                && outbuf[j] == '\r' && outbuf[j + 1] == '\n')
                break;
            break;
          case 1:
            for (j = i; j < length; j++)
              if (outbuf[j] == '\r')
                break;
            break;
          default:
            for (j = i; j < length; j++)
              if (outbuf[j] == '\n')
                break;
        }
      p->length = j - i;
      p->text = g_malloc (p->length * sizeof (gchar));
      g_memmove (p->text, outbuf + i, p->length);
      if (j < length)
        {
          LineBuffer *q;

          j += tmaid->ft.crlf == 0 ? 2 : 1;
          q = g_malloc0 (sizeof (LineBuffer));
          q->prev = p;
          q->next = p->next;
          p->next = q;
          p = q;
          tmaid->max++;
        }
      i = j;
    }
  g_free (outbuf);
  /* ja:変換 */
  if (tmaid->ft.html)
    for (p = tmaid->start; p; p = p->next)
      {
        while (gtk_events_pending ())
          gtk_main_iteration ();
        i = 0;
        while (i + 4 <= p->length)
          if (p->text[i++] == '&' && p->text[i++] == '#')
            {
              c = 0;
              if (g_ascii_tolower (p->text[i]) == 'x')
                {
                  /* ja:16進数 */
                  for (j = i + 1; j < p->length; j++)
                    if (g_ascii_isxdigit (p->text[j]))
                      c = c * 16 + g_ascii_xdigit_value (p->text[j]);
                    else
                      break;
                }
              else
                {
                  /* ja:10進数 */
                  for (j = i; j < p->length; j++)
                    if (g_ascii_isdigit (p->text[j]))
                      c = c * 10 + g_ascii_digit_value (p->text[j]);
                    else
                      break;
                }
              if (i < j && j < p->length && p->text[j++] == ';')
                {
                  i -= 2;
                  leng = g_unichar_to_utf8 (c, buf);
                  if (leng > j - i)
                    p->text = g_realloc (p->text, p->length + leng - (j - i));
                  g_memmove (p->text + i + leng, p->text + j, p->length - j);
                  g_memmove (p->text + i, buf, leng);
                  p->length = p->length + leng - (j - i);
                  i += leng;
                }
            }
      }
  else if (tmaid->ft.esc)
    for (p = tmaid->start; p; p = p->next)
      {
        while (gtk_events_pending ())
          gtk_main_iteration ();
        i = 0;
        while (i + 6 <= p->length)
          if (p->text[i++] == '\\' && p->text[i++] == 'u')
            {
              c = 0;
              for (j = i; j < p->length; j++)
                if (g_ascii_isxdigit (p->text[j]))
                  c = c * 16 + g_ascii_xdigit_value (p->text[j]);
                else
                  break;
              if (i + 3 == j)
                {
                  i -= 2;
                  leng = g_unichar_to_utf8 (c, buf);
                  if (leng > j - i)
                    p->text = g_realloc (p->text, p->length + leng - (j - i));
                  g_memmove (p->text + i + leng, p->text + j, p->length - j);
                  g_memmove (p->text + i, buf, leng);
                  p->length = p->length + leng - (j - i);
                  i += leng;
                }
            }
      }
  /* ja:中断ボタン */
  abort_box_end ();
}


/*  ja:ファイルを開く
    file,ファイル名
     RET,ウインドウ情報                                                     */
TmaidWindow *
file_open_edit (const gchar *file)
{
  gchar *filename;
  gint i, page_num;
  TmaidWindow *tmaid;
  GdkRectangle rc;
  GtkWidget *table;
#ifndef G_OS_WIN32
  gint x, y;
  GtkWidget *preedit;
#endif /* not G_OS_WIN32 */

  tmaid = g_malloc0 (sizeof (TmaidWindow));
  tmaid->select.x = -1;
  if (!file)
    {
      gchar **charset;
      static gint fcount = 0;   /* ja:新規ファイルの数 */

      /* ja:新規 */
      filename = g_strdup_printf ("new%04d", fcount++ % 1000);
      property_copy (&tmaid->ft, ftype);
      if (!ftype[0].associate)
        tmaid->ft.ft_id = -1;
#ifdef G_OS_WIN32
      tmaid->ft.crlf = ftype[0].crlf % 3;
#else /* not G_OS_WIN32 */
      tmaid->ft.crlf = MIN (ftype[0].crlf, 2);
#endif /* not G_OS_WIN32 */
      charset = g_strsplit (ftype[0].charset, ",", 0);
      tmaid->ft.charset = g_strdup (charset[0]);
      g_strfreev (charset);
      /* ja:ファイルが0バイトのとき */
      tmaid->max = 1;
      tmaid->start = g_malloc0 (sizeof (LineBuffer));
    }
  else
    {
      gchar *name, *text;

      filename = fileio_get_full_path (file);
      name = g_path_get_basename (file);
      text = g_filename_to_utf8 (name, -1, NULL, NULL, NULL);
      g_free (name);
      for (i = 0; i < ftnum; i++)
        {
          gchar *ext, *p;

          ext = g_strdup (ftype[i].ext);
          for (p = g_strtok (ext, G_SEARCHPATH_SEPARATOR_S); p;
                                p = g_strtok (NULL, G_SEARCHPATH_SEPARATOR_S))
            if (g_pattern_match_simple (p, text))
              goto loop;
          g_free (ext);
        }
      loop:
      g_free (text);
      if (i >= ftnum)
      i = 0;
      property_copy (&tmaid->ft, ftype + i);
      if (!ftype[i].associate)
        tmaid->ft.ft_id = -1;
      tmaid->ft.crlf = ftype[i].crlf;
      tmaid->ft.charset = g_strdup (ftype[i].charset);
      file_open_text (file, tmaid);
    }

  /* ja:Input Method */
  tmaid->im_context = gtk_im_multicontext_new ();
  g_signal_connect (tmaid->im_context, "commit",
                            G_CALLBACK (signal_commit), tmaid);
#ifndef G_OS_WIN32
  g_signal_connect (tmaid->im_context, "preedit-changed",
                            G_CALLBACK (signal_preedit_changed), tmaid);
#endif /* not G_OS_WIN32 */
  g_signal_connect (tmaid->im_context, "retrieve-surrounding",
                            G_CALLBACK (signal_retrieve_surrounding), tmaid);
  g_signal_connect (tmaid->im_context, "delete-surrounding",
                            G_CALLBACK (signal_delete_surrounding), tmaid);
#ifdef G_OS_WIN32
  gtk_im_context_set_use_preedit (tmaid->im_context, FALSE);
#else /* not G_OS_WIN32 */
  preedit = gtk_window_new (GTK_WINDOW_POPUP);
  tmaid->preedit = gtk_drawing_area_new ();
  g_signal_connect (G_OBJECT (tmaid->preedit), "expose-event",
                                    G_CALLBACK (signal_expose_preedit), tmaid);
  gtk_container_add (GTK_CONTAINER (preedit), tmaid->preedit);
  gtk_widget_show (tmaid->preedit);
#endif /* not G_OS_WIN32 */

  /* ja:表示エリア */
  tmaid->drawing = gtk_drawing_area_new ();
  g_signal_connect (G_OBJECT (tmaid->drawing), "focus-in-event",
                                    G_CALLBACK (signal_focus_in), tmaid);
  g_signal_connect (G_OBJECT (tmaid->drawing), "focus-out-event",
                                    G_CALLBACK (signal_focus_out),tmaid);
  g_signal_connect (G_OBJECT (tmaid->drawing), "realize",
                                    G_CALLBACK (signal_realize), tmaid);
  g_signal_connect (G_OBJECT (tmaid->drawing), "unrealize",
                                    G_CALLBACK (signal_unrealize), tmaid);
  g_signal_connect (G_OBJECT (tmaid->drawing), "configure-event",
                                    G_CALLBACK (signal_config), tmaid);
  g_signal_connect (G_OBJECT (tmaid->drawing), "expose-event",
                                    G_CALLBACK (signal_expose), tmaid);
  g_signal_connect (G_OBJECT (tmaid->drawing), "button-press-event",
                                    G_CALLBACK (signal_button_press), tmaid);
  g_signal_connect (G_OBJECT (tmaid->drawing), "motion-notify-event",
                                    G_CALLBACK (signal_motion_notify), tmaid);
  g_signal_connect (G_OBJECT (tmaid->drawing), "button-release-event",
                                    G_CALLBACK (signal_button_release), tmaid);
  g_signal_connect (G_OBJECT (tmaid->drawing), "scroll-event",
                                    G_CALLBACK (signal_scroll), tmaid);
  g_signal_connect_after (G_OBJECT (tmaid->drawing), "key-press-event",
                                    G_CALLBACK (signal_key_press), tmaid);
  g_signal_connect_after (G_OBJECT (tmaid->drawing), "key-release-event",
                                    G_CALLBACK (signal_key_release), tmaid);
  g_signal_connect (G_OBJECT (tmaid->drawing), "destroy",
                                    G_CALLBACK (signal_destroy_draw), tmaid);
  gtk_widget_add_events (tmaid->drawing, GDK_POINTER_MOTION_MASK
                            | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
                            | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK
                            | GDK_FOCUS_CHANGE_MASK | GDK_SCROLL_MASK);
  GTK_WIDGET_SET_FLAGS (tmaid->drawing, GTK_CAN_FOCUS);
  /* ja:フォント */
  if (tmaid->ft.font_name)
    {
      PangoFontDescription *desc;

      desc = pango_font_description_from_string (tmaid->ft.font_name);
      gtk_widget_modify_font (window, desc);
      pango_font_description_free (desc);
    }
  else
    {
      gtk_widget_modify_font (window, system_font);
    }
  tmaid->layout = gtk_widget_create_pango_layout (window, NULL);
  charset_size (tmaid->layout, &tmaid->font_width, &tmaid->font_height,
                                                            tmaid->font_buf);
  /* ja:プライマリセレクション */
  gtk_selection_add_targets (tmaid->drawing, GDK_SELECTION_PRIMARY,
                                            targets, G_N_ELEMENTS (targets));
  g_signal_connect (G_OBJECT (tmaid->drawing), "selection-get",
                                    G_CALLBACK (signal_selection_get), tmaid);
  /* ja:スクロールバー */
  tmaid->hscroll = gtk_hscrollbar_new
                    (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 0, 1, 1, 1)));
  tmaid->vscroll = gtk_vscrollbar_new
                    (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 0, 1, 1, 1)));
  /* ja:テーブル */
  table = gtk_table_new (2, 2, FALSE);
  gtk_table_attach (GTK_TABLE (table), tmaid->drawing, 0, 1, 0, 1,
                                    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                                    GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  gtk_table_attach (GTK_TABLE (table), tmaid->hscroll, 0, 1, 1, 2,
                            GTK_EXPAND | GTK_SHRINK | GTK_FILL,GTK_FILL, 0, 0);
  gtk_table_attach (GTK_TABLE (table), tmaid->vscroll, 1, 2, 0, 1,
                            GTK_FILL,GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  g_object_set_data (G_OBJECT(table), "user_data", tmaid);
  /* ja:表示 */
  page_num = orz_mdi_append_page (ORZ_MDI (mdi),
                                                table, filename, !file, tmaid);
  gtk_widget_show_all (table);
  gtk_notebook_set_current_page (GTK_NOTEBOOK (mdi), page_num);
  gtk_widget_grab_focus (tmaid->drawing);

  /* ja:Input Method */
  while (gtk_events_pending ())
    gtk_main_iteration ();
#ifndef G_OS_WIN32
  gdk_window_get_origin (tmaid->drawing->window, &x, &y);
  gtk_window_move (GTK_WINDOW (gtk_widget_get_parent (tmaid->preedit)), x, y);
#endif /* not G_OS_WIN32 */
  rc.x = rc.y = rc.width = 0;
  rc.height = tmaid->font_height;
  gtk_im_context_set_cursor_location (tmaid->im_context, &rc);
  /* ja:新規作成ではなく、ファイルが存在する時には履歴に加える */
  if (file && g_file_test (file, G_FILE_TEST_EXISTS))
    orz_history_add_file (ORZ_HISTORY (history), file);
  g_free (filename);
  /* ja:右マージン */
  if (tmaid->ft.limit)
    modify_margin (tmaid);

  return tmaid;
}


/******************************************************************************
*                                                                             *
* ja:ファイル出力関数群                                                       *
*                                                                             *
******************************************************************************/
/*  TXTファイルを保存する
     file,ファイル名
    tmaid,ウインドウ情報
      RET,TRUE:正常終了,FALSE:エラー                                        */
gboolean
file_save_text (const gchar *file,
                TmaidWindow *tmaid)
{
  gchar *buf = NULL, crlf[2], *ptr;
  gsize crlf_leng, length = 0;
  FileIO *fio;
  LineBuffer *p;
  GError *error = NULL;

  p = tmaid->start;
  while (p->prev)
    p = p->prev;
  switch (tmaid->ft.crlf)
    {
      case  0: crlf[0] = '\r'; crlf[1] = '\n'; crlf_leng = 2; break;
      case  1: crlf[0] = '\r'; crlf_leng = 1; break;
      default: crlf[0] = '\n'; crlf_leng = 1;
    }
  /* ja:中断ボタン */
  abort_box_start ();
  /* ja:書く */
  while (abort_break && p)
    {
      gsize leng;

      if (!p->margin && p->next)
        {
          leng = p->length + crlf_leng;
          p->text = realloc (p->text, leng);
          g_memmove (p->text + p->length, crlf, crlf_leng);
        }
      else if (!p->next && tmaid->ft.eof && p->length > 0
                                        && p->text[p->length - 1] != '\x1a')
        {
          leng = p->length + 1;
          p->text = realloc (p->text, leng);
          p->text[p->length] = '\x1a';
        }
      else
        {
          leng = p->length;
        }
      ptr = p->text;
      while (abort_break && leng > 0)
        {
          gchar *outbuf;
          gsize bytes_read, bytes_written;
          gssize bytes_read_prev;

          while (gtk_events_pending ())
            gtk_main_iteration ();
          bytes_read_prev = leng;
          while ((g_clear_error (&error),
                                !(outbuf =  g_convert (ptr, bytes_read_prev,
                                        tmaid->ft.charset, "UTF-8",
                                        &bytes_read, &bytes_written, &error)))
                            && bytes_read_prev != bytes_read && bytes_read > 0)
            bytes_read_prev = bytes_read;
          if (bytes_read > 0)
            {
              if (outbuf && bytes_written > 0)
                {
                  buf = realloc (buf, length + bytes_written);
                  g_memmove (buf + length, outbuf, bytes_written);
                  length += bytes_written;
                }
              leng -= bytes_read;
              ptr += bytes_read;
            }
          g_free (outbuf);
          if (!outbuf && !g_error_matches (error, G_CONVERT_ERROR,
                                            G_CONVERT_ERROR_ILLEGAL_SEQUENCE)
                                && !g_error_matches (error, G_CONVERT_ERROR,
                                                G_CONVERT_ERROR_PARTIAL_INPUT))
            {
              /* ja:キャラクターセットが無効 */
              GtkWidget *dialog;

              g_clear_error (&error);
              g_free (buf);
              /* ja:中断ボタン */
              abort_box_end ();
              dialog = gtk_message_dialog_new (GTK_WINDOW (window),
                    GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                    GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
                    _("Conversion from character encoding"
                    " 'UTF-8' to '%s' is not supported"), tmaid->ft.charset);
              g_signal_connect (G_OBJECT (dialog), "key-press-event",
                                    G_CALLBACK (misc_dialog_key_press), NULL);
              gtk_dialog_run (GTK_DIALOG (dialog));
              gtk_widget_destroy (dialog);
              return FALSE;
            }
          if (bytes_read <= 0)
            {
              switch (tmaid->ft.writemode)
                {
                  case 1:/* '?' */
                    outbuf = g_strdup ("?");
                    break;
                  case 2:/* HTML(10進数) */
                    outbuf = g_strdup_printf ("&#%d;", g_utf8_get_char (ptr));
                    break;
                  case 3:/* HTML(16進数) */
                    outbuf = g_strdup_printf (tmaid->ft.hexcase
                                ? "&#x%X;" : "&#x%x;", g_utf8_get_char (ptr));
                    break;
                  case 4:/* Escape */
                    outbuf = g_strdup_printf (tmaid->ft.hexcase
                            ? "\\u%04X" : "\\u%04x", g_utf8_get_char (ptr));
                    break;
                  default:
                    {
                      GtkWidget *dialog;

                      g_free (buf);
                      /* ja:中断ボタン */
                      abort_box_end ();
                      dialog = gtk_message_dialog_new (GTK_WINDOW (window),
                            GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                            GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
                            _("Invalid byte sequence in conversion input"));
                      g_signal_connect (G_OBJECT (dialog), "key-press-event",
                                    G_CALLBACK (misc_dialog_key_press), NULL);
                      gtk_dialog_run (GTK_DIALOG (dialog));
                      gtk_widget_destroy (dialog);
                      return FALSE;
                    }
                }
              bytes_written = g_strlen (outbuf);
              buf = realloc (buf, length + bytes_written);
              g_memmove (buf + length, outbuf, bytes_written);
              length += bytes_written;
              leng -= charset_length (*ptr);
              ptr += charset_length (*ptr);
              g_free (outbuf);
            }
        }
      p = p->next;
    }
  g_clear_error (&error);
  if (abort_break)
    {
      /* ja:開く */
      fio = fileio_open (file, FILEIO_ACCESS_WRITE, FILEIO_SHARE_READ,
                                                    FILEIO_MODE_CREATE_ALWAYS);
      if (!fio)
        {
          GtkWidget *dialog;

          g_free (buf);
          /* ja:中断ボタン */
          abort_box_end ();
          dialog = gtk_message_dialog_new (GTK_WINDOW (window),
                            GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                            GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
                            _("Can not open file"));
          g_signal_connect (G_OBJECT (dialog), "key-press-event",
                                    G_CALLBACK (misc_dialog_key_press), NULL);
          gtk_dialog_run (GTK_DIALOG (dialog));
          gtk_widget_destroy (dialog);
          return FALSE;
        }
      ptr = buf;
      while (abort_break && length > 0)
        {
          gssize leng;

          while (gtk_events_pending ())
            gtk_main_iteration ();
          leng = MIN (length, 16384);
          if (fileio_write (fio, ptr, leng) != leng)
            {
              GtkWidget *dialog;

              fileio_close (fio);
              g_free (buf);
              /* ja:中断ボタン */
              abort_box_end ();
              dialog = gtk_message_dialog_new (GTK_WINDOW (window),
                            GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                            GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
                            _("Can not write"));
              g_signal_connect (G_OBJECT (dialog), "key-press-event",
                                    G_CALLBACK (misc_dialog_key_press), NULL);
              gtk_dialog_run (GTK_DIALOG (dialog));
              gtk_widget_destroy (dialog);
              return FALSE;
            }
          length -= leng;
          ptr += leng;
        }
      if (!fileio_close (fio))
        {
          GtkWidget *dialog;

          g_free (buf);
          /* ja:中断ボタン */
          abort_box_end ();
          dialog = gtk_message_dialog_new (GTK_WINDOW (window),
                            GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                            GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
                            "fileio_close");
          g_signal_connect (G_OBJECT (dialog), "key-press-event",
                                    G_CALLBACK (misc_dialog_key_press), NULL);
          gtk_dialog_run (GTK_DIALOG (dialog));
          gtk_widget_destroy (dialog);
          return FALSE;
        }
    }
  g_free (buf);
  /* ja:中断ボタン */
  abort_box_end ();
  return abort_break;
}
